library rx_property;

import 'dart:async';
import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart';

typedef Combiner<A, B, T> = T Function(A, B);

typedef Observer<S, T> = S Function(T);

/// An abstraction layer for RxDart Subjects.
/// Reactive example:
/// ```
/// class RxCounter extends Reactive<int> {
///   RxCounter() : super(0);
///
///   void increment() => value++;
///   void decrement() => value--;
/// }
///```
mixin ReactiveMixin<T> {
  // The internel stream controller
  Subject<T> _subject;

  /// Return whether the controller has already been closed or not.
  /// Once closed, it is not allowed to change the value.
  ///
  /// If add() or the set value is used after the stream is closed an exception will be thrown.
  bool isClosed = false;

  bool _distinctUntilChanged = false;

  T _lastValue;

  /// Returns the latest emited value by the Reactive.
  T get value => _lastValue;

  /// Sends a new data [value].
  /// Is equivalent to add()
  @protected
  set value(T value) => _setValue(value);

  /// Sends a new data [value].
  @protected
  void _setValue(T value) async {
    if (_distinctUntilChanged && value == _lastValue) return;
    _lastValue = value;
    _subject.add(_lastValue);
  }

  void addError(Object error, [StackTrace stackTrace]) {
    _subject.addError(error, stackTrace);
  }

  /// Closes the sink.
  /// The [add] method and use the set [value] must not be called after this method.
  /// Calling this method more than once is allowed, but does nothing.
  Future<void> close() async {
    if (isClosed) return;
    isClosed = true;
    await _subject.close();
  }

  @override
  String toString() {
    return _lastValue == null ? 'Reactive $T null value.' : '$_lastValue';
  }

  // Sends the current value.
  void notify() => _setValue(_lastValue);
}

Stream<T> computed<A, B, T>(
    Stream<A> streamA, Stream<B> streamB, T Function(A a, B b) combiner) {
  return CombineLatestStream.combine2(streamA, streamB, combiner);
}

Stream<S> observable<T, S>(Reactive<T> reactive, S Function(T) convert) {
  return reactive.map<S>(convert);
}

abstract class ReactiveController<T> extends StreamView<T>
    with ReactiveMixin<T>
    implements EventSink<T> {
  ReactiveController(T initalValue, Subject<T> subject)
      : super(subject.stream) {
    _subject = subject;
    _lastValue = initalValue;
  }

  @protected
  void add(T event) => _setValue(event);
}

abstract class Reactive<T> extends ReactiveController<T> {
  Reactive(T initalValue)
      : super(initalValue, BehaviorSubject.seeded(initalValue));

  void change(T Function(T value) changer) {
    add(changer(value));
  }

  static bool distinctUntilChanged = true;
  bool _distinctUntilChanged = distinctUntilChanged;
}

abstract class ReplayReactive<T> extends ReactiveController<T> {
  ReplayReactive(T initalValue) : super(initalValue, ReplaySubject());
}

class PublishReactive<T> extends ReactiveController<T> {
  PublishReactive(T initalValue) : super(initalValue, PublishSubject());
}

class ReactiveProperty<T> extends Reactive<T> {
  ReactiveProperty(T initalValue) : super(initalValue);
  @override
  set value(T value) {
    super.value = value;
  }

  @override
  add(T value) {
    super.add(value);
  }

  static ReplayReactiveProperty<T> replay<T>(T initalValue) {
    return ReplayReactiveProperty(initalValue);
  }

  static ReactiveProperty<T> publish<T>(T initalValue) {
    return PublishReactiveProperty(initalValue);
  }
}

class ReplayReactiveProperty<T> extends ReplayReactive<T> {
  void change(T Function(T value) changer) {
    add(changer(value));
  }

  ReplayReactiveProperty(T initalValue) : super(initalValue);
}

class PublishReactiveProperty<T> extends ReactiveProperty<T> {
  PublishReactiveProperty(T initalValue) : super(initalValue);
}

extension ReactiveExtension<T> on Stream<T> {
  Stream<R> comput<S, R>(Stream<S> latestFromStream, R Function(T, S) convert) {
    return CombineLatestStream.combine2(this, latestFromStream, convert);
  }

  Stream<S> observe<S>(Observer<S, T> observe) {
    return this.map(observe);
  }
}
